/* Weckkubator - Weck(topf)-(In)kubator
 * 
 * Version 0.1.0
 * 
 * Fliegerjohn - 06.05.2024
 */



#include <Wire.h>  // Support for I2C --> LCD
#include <LiquidCrystal_I2C.h>  // Support for PCF8574 --> LCD
#include <OneWire.h>  // 1-Wiresupport for DS18B20
#include <DallasTemperature.h>  // Easy reading of DS18B20
#include <PID_v1.h>  // PID-Controller

#define ONE_WIRE_BUS 10  // Enable 1-Wirebus on digital 10 (PB2 - Pin16)
#define fan 8
#define wecktopf 9

LiquidCrystal_I2C lcd(0x3F,20,4);  // declare LCD
OneWire oneWire(ONE_WIRE_BUS);  // declare 1-Wirebus
DallasTemperature sensors(&oneWire);  // declare sensors 

// Define variables PID is connected to
double pidset, watertemp, output;

// PID adjustment
double Kp=500, Ki=3, Kd=1;
PID myPID(&watertemp, &output, &pidset, Kp, Ki, Kd, DIRECT);
int WindowSize = 5000;
unsigned long windowStartTime;

// Temp adjustment
float tempset;
int poti;

// time
unsigned long previousMillis;
unsigned long previousMillisFan;


// °C for LCD
byte customChar[8] = {
  0b01110,
  0b01010,
  0b01010,
  0b01110,
  0b00000,
  0b00000,
  0b00000,
  0b00000
};

void setup() {
  
  // Setup GPIO
  pinMode(wecktopf, OUTPUT);  // Output "Wecktopf"
  pinMode(fan, OUTPUT);  // Fan
  
  // Setup sensors
  sensors.begin();
  
  // Setup I2C-LCD
  lcd.init();
  lcd.backlight();
  lcd.clear();
  lcd.createChar(0, customChar);
  lcd.setCursor(1, 0);
  
  // Setup PID
  windowStartTime = millis();
  myPID.SetOutputLimits(0, WindowSize);  //tell the PID to range between 0 and the full window size
  myPID.SetMode(AUTOMATIC);  //turn on the PID
  
  // Greetings
  lcd.print("Weckkubator 0.1.0");
  delay(3000);


  // Setup LCD
  setupdisplay();

}

void loop() {

  // run pid
  pid();
  
  // update LCD
  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis >= 250) {
    settemp();
    readtemp();
    lcdoutput();
    previousMillis = currentMillis;
  }

  // start fan if heating on full power
  if (currentMillis - previousMillisFan >= 10000) {
    if (output >= 4800) digitalWrite(fan, HIGH);
    else digitalWrite(fan, LOW);
    previousMillisFan = currentMillis;
  }
}

//------------------------------------------
// PID-Controller
void pid() {

  myPID.Compute();
  
  if (millis() - windowStartTime > WindowSize)
  { //time to shift the Relay Window
    windowStartTime += WindowSize;
  }
  if (output < millis() - windowStartTime) digitalWrite(wecktopf, LOW);
  else digitalWrite(wecktopf, HIGH);
}

//------------------------------------------
// Read temp from DS18B20
void readtemp() {

  /* Read temp of first sensor and take if for watertemp.
   * In future we might add others.
   */
  sensors.requestTemperatures();
  watertemp = sensors.getTempCByIndex(0);

  if (watertemp == -127.00) missingsensor();
  
}

//------------------------------------------
// LCD output
void lcdoutput() {
  // Output temperatureset
  lcd.setCursor(8, 1);
  lcd.print(pidset);
  lcd.write((byte)0);
  lcd.print("C");

  // Output watertemp
  lcd.setCursor(8, 2);
  lcd.print(watertemp);
  lcd.write((byte)0);
  lcd.print("C");
  lcd.print("   ");

  // output PID
  lcd.setCursor (8, 3);
  lcd.print(output);
  lcd.print("   ");

}


//------------------------------------------
// Read Poti
void settemp() {
  poti = analogRead(A3);
  tempset = (20+(poti*0.04));

  if (tempset <= 50) {
    pidset = tempset;
  }

  if (tempset >= 51) {
    pidset = (tempset+30);
  }

  
}
//------------------------------------------
// Missing Sensor
void missingsensor() {
  // turn of all output
  digitalWrite(wecktopf, LOW);
  digitalWrite(fan, LOW);

  // Clear LCD
    lcd.clear();
    
  do {
    lcd.setCursor(0, 0);
    lcd.print("Error:");
    lcd.setCursor(0, 1);
    lcd.print("Missing Sensor");
    sensors.requestTemperatures();
    watertemp = sensors.getTempCByIndex(0);
    } while (watertemp == -127.00);

  setupdisplay();
}

//------------------------------------------
// Setup LCD
void setupdisplay() {

  // Clear LCD
  lcd.clear();
  
  // Output temperatureset
  lcd.setCursor(0, 1);
  lcd.print("set:    ");
  lcd.print(pidset);
  lcd.write((byte)0);

  // Output watertemp
  lcd.setCursor(0, 2);
  lcd.print("temp:   ");
  lcd.print(watertemp);
  lcd.write((byte)0);
  lcd.print("C");

  // output PID
  lcd.setCursor (0, 3);
  lcd.print("output: ");
  lcd.print(output);
}
